/****************************************************************************
 Include Section
*****************************************************************************/

#include "global.h"
#include "INIT.h"
#include "data.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

#if defined SPI_DMA
    #define SPI_CTL_VAL_TX  TDBR_DMA
    #define SPI_CTL_VAL_RX  RDBR_DMA
#else
    #define SPI_CTL_VAL_TX  RDBR_CORE
    #define SPI_CTL_VAL_RX  RDBR_CORE
#endif


#define DMA7_CONFIG_VAL     FLOW_STOP
#define DMA7_X_COUNT_VAL    N
#if defined SPI_WDSIZE_16
    #define DMA7_X_MODIFY_VAL   2
#else
    #define DMA7_X_MODIFY_VAL   1
#endif
#define DMA7_Y_COUNT_VAL    0
#define DMA7_Y_MODIFY_VAL   0

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __spi_run_tx;
.GLOBAL __spi_run_rx;
.GLOBAL __spi_stop;
 
#if defined PFHW
    .GLOBAL __hostwait_check;
    .GLOBAL __hostwait_assert;
#endif

//.GLOBAL spi_core_loop_run_tx;
//.GLOBAL spi_core_loop_run_rx;
//.GLOBAL spi_init_tx;
//.GLOBAL spi_init_rx;
//.GLOBAL spi_init;
//.GLOBAL spi_start;
//.GLOBAL spi_putc;
//.GLOBAL spi_getc;
//.GLOBAL spi_calc_baud;
//.GLOBAL hostwait_deassert;

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION L1_code;

/*****************************************************************************
 Function:              __spi_run_tx
 Description:           run a SPI transmission -> receive SPI
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__spi_run_tx:
    link 0;

    call spi_init_tx;
    call spi_start;
    #if defined SPI_CORE_LOOP
        call spi_core_loop_run_tx;
    #endif

    unlink;
    rts;
__spi_run_tx.END:


/*****************************************************************************
 Function:              __spi_run_rx
 Description:           run a SPI tranmission -> transmit SPI
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        none
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__spi_run_rx:
    link 0;

    call spi_init_rx;
    call spi_start;
    #if defined SPI_CORE_LOOP
        call spi_core_loop_run_rx;
    #endif

    unlink;
    rts;
__spi_run_rx.END:


/*****************************************************************************
 Function:              spi_core_loop_run_tx
 Description:           run a SPI tranmission -> transmit SPI
                        Hardware Loop used
                        No Interrupts
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,R0,P5:3,ASTAT,LOOPREG
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_core_loop_run_tx:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:3);
    [--SP] = R0;
    PUSH_LOOP_REGx(0);
    IMM32(P5,SYS_MMR_BASE);


    /*************
    * SPI Enable *
    *************/
    call spi_init_tx;
    call spi_start;

    /***************
    * SPI transmit *
    ***************/
    IMM32(P4,tx_buf);
    IMM32(P3,N);

    /***********************
    * SPI-Transfer Section *
    ***********************/
    lsetup( spi_core_loop_run_tx.loop_start, spi_core_loop_run_tx.loop_end ) LC0 = P3;
    spi_core_loop_run_tx.loop_start:
    #if defined SPI_WDSIZE_16
        R0 = w[P4++] (z); /* Get next data to be transmitted */
    #else
        R0 = b[P4++] (z); /* Get next data to be transmitted */
    #endif
    spi_core_loop_run_tx.loop_end: call spi_putc;

    /**************
    * SPI Disable *
    **************/
    call __spi_stop;


    POP_LOOP_REGx(0);
    R0 = [SP++];
    (R7:7,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_core_loop_run_tx.end:


/*****************************************************************************
 Function:              spi_putc
 Description:           run a SPI tranmission -> transmit SPI
                        Hardware Loop used
                        No Interrupts
 Input Parameters:      R0
 Return Parameters:     none
 Registers Used:        R7:6,P5,ASTAT
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_putc:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    w[P5 + lo(SPI_TDBR)] = R0; /* Write to SPI_TDBR */

    #if defined SPI_MASTER_MODE
        R7 = w[P5 + lo(SPI_CTL)] (z);
        R6 = TIMOD (z);
        R7 = R7 & R6;
        CC = R7 == RDBR_CORE;
        if !CC jump spi_putc.check_txs;
       /* Dummy read of SPI_RDBR kicks off transfer */
        R7 = w[P5 + lo(SPI_RDBR)] (z);
    #endif

    spi_putc.check_txs:     R7 = w[P5 + lo(SPI_STAT)] (z);
                            CC = bittst (R7,bitpos(TXS));
                            if CC jump spi_putc.check_txs;

    /* Poll for SPIF = 1 (SPI finished) */
    spi_putc.check_spif:    R7 = w[P5 + lo(SPI_STAT)] (z);
                            CC = bittst (R7,bitpos(SPIF));
                            if !CC jump spi_putc.check_spif;


    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_putc.end:


/*****************************************************************************
 Function:              spi_core_loop_run_rx
 Description:           run a SPI tranmission -> transmit SPI
                        Hardware Loop used
                        No Interrupts
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,R0,P5:3,ASTAT,LOOPREG
 Global Variables Used: none
 C-Callable :           no
 HINT:                  SPI Master Mode Version
*****************************************************************************/

spi_core_loop_run_rx:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:3);
    [--SP] = R0;
    PUSH_LOOP_REGx(0);
    IMM32(P5,SYS_MMR_BASE);


    /*************
    * SPI Enable *
    *************/
    call spi_init_rx;
    call spi_start;

    /**************
    * SPI Receive *
    **************/
    IMM32(P4,rx_buf);
    #if defined SPI_MASTER_MODE
        IMM32(P3,(N-1));
    #elif defined SPI_SLAVE_MODE
        IMM32(P3,N);
    #endif

    /***********************
    * Pre-Transfer Section *
    ***********************/
    #if defined SPI_MASTER_MODE
        R7 = w[P5 + lo(SPI_RDBR)] (z); /* first read, dummy read */
    #endif

    /***********************
    * SPI-Transfer Section *
    ***********************/
    lsetup( spi_core_loop_run_rx.loop_start, spi_core_loop_run_rx.loop_end ) LC0 = P3;
    spi_core_loop_run_rx.loop_start: call spi_getc;

    #if defined SPI_WDSIZE_16
        spi_core_loop_run_rx.loop_end: w[P4++] = R0;
    #else
        spi_core_loop_run_rx.loop_end: b[P4++] = R0;
    #endif

    /************************
    * Post-Transfer Section *
    ************************/
    #if defined SPI_MASTER_MODE
        /* two consecutive reads of RXS are necessary */
        spi_core_loop_run_rx.check_rxs_last: /* Poll for RXS = 0 (Receive Data Buffer empty) */
        R7 = w[P5 + lo(SPI_STAT)] (z);
        CC = bittst (R7,bitpos(RXS));
        if !CC jump spi_core_loop_run_rx.check_rxs_last;
        R7 = w[P5 + lo(SPI_STAT)] (z);
        CC = bittst (R7,bitpos(RXS));
        if !CC jump spi_core_loop_run_rx.check_rxs_last;

        /* Poll for SPIF = 1 (SPI finished) */
        spi_core_loop_run_rx.check_spif_last:
        R7 = w[P5 + lo(SPI_STAT)] (z);
        CC = bittst (R7,bitpos(SPIF));
        if !CC jump spi_core_loop_run_rx.check_spif_last;

        R7 = w[P5 + lo(SPI_SHADOW)] (z);
        #if defined SPI_WDSIZE_16
            w[P4] = R7; /* last word to be copied from SPI_SHADOW (=SPI_RDBR) to memory */
        #else
            b[P4] = R7; /* last word to be copied from SPI_SHADOW (=SPI_RDBR) to memory */
        #endif
    #endif

    /**************
    * SPI Disable *
    **************/
    call __spi_stop;


    POP_LOOP_REGx(0);
    R0 = [SP++];
    (R7:7,P5:3) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_core_loop_run_rx.end:


/*****************************************************************************
 Function:              spi_getc
 Description:           run a SPI tranmission -> transmit SPI
                        Hardware Loop used
                        No Interrupts
 Input Parameters:      none
 Return Parameters:     R0
 Registers Used:        R7,P5,ASTAT
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_getc:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* two consecutive reads of RXS are necessary */
    spi_getc.check_rxs: R7 = w[P5 + lo(SPI_STAT)] (z);
                        CC = bittst (R7,bitpos(RXS));
                        if !CC jump spi_getc.check_rxs;
                        R7 = w[P5 + lo(SPI_STAT)] (z);
                        CC = bittst (R7,bitpos(RXS));
                        if !CC jump spi_getc.check_rxs;

    /* Poll for SPIF = 1 (SPI finished) */
    spi_getc.check_spif:    R7 = w[P5 + lo(SPI_STAT)] (z);
                            CC = bittst (R7,bitpos(SPIF));
                            if !CC jump spi_getc.check_spif;

    /* read word from SPI_RDBR */
    /* for a SPI master device this read from SPI_RDBR is requesting a next word from a slave device */
    /* therefore SPI_SHADOW should be read for the very last word */
    R0 = w[P5 + lo(SPI_RDBR)] (z);


    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_getc.end:


/*****************************************************************************
 Function:              spi_init_tx
 Description:           Initialize SPI for transmiting
                        Initialize DMA for transmiting if defined
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_init_tx:

    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /******
    * SPI *
    ******/
    R7 = SPI_CTL_VAL_TX (z);
    #if defined SPI_SZ
        bitset(R7,bitpos(SZ));
    #endif
    /* use EMISO bit to identify transmit or receive mode for both master and slave */
    bitset(R7,bitpos(EMISO));
    w[P5 + lo(SPI_CTL)] = R7;


    /******
    * DMA *
    ******/
    #if defined SPI_DMA
        R7 =    FLOW_STOP   |   /* Next Operation */
                DI_EN       |   /* Data Interrupt Enable */
               nDI_SEL      |   /* Data Interrupt Timing Select */
               nSYNC        |   /* Work Unit Transitions */
               nDMA2D       |   /* DMA Mode */
                WDSIZE_8    |   /* Transfer Word Size */
               nWNR         |   /* DMA Direction */
               nDMAEN       (z);/* DMA Channel Enable */
        w[P5 + lo(DMA7_CONFIG)] = R7;

        IMM32(R7,tx_buf);
        [P5 + lo(DMA7_START_ADDR)] = R7;
    #endif

    call spi_init;


    (R7:7,P5:5) = [SP++];
    unlink;

    rts;
spi_init_tx.end:


/*****************************************************************************
 Function:              spi_init_rx
 Description:           Initialize SPI for receiving
                        Initialize DMA for receiving if defined
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_init_rx:

    link 0;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /******
    * SPI *
    ******/
    R7 = SPI_CTL_VAL_RX (z);
    #if defined SPI_GM
        bitset(R7,bitpos(GM));
    #endif
    w[P5 + lo(SPI_CTL)] = R7;


    /******
    * DMA *
    ******/
    #if defined SPI_DMA
        R7 =    FLOW_STOP   |   /* Next Operation */
                DI_EN       |   /* Data Interrupt Enable */
               nDI_SEL      |   /* Data Interrupt Timing Select */
               nSYNC        |   /* Work Unit Transitions */
               nDMA2D       |   /* DMA Mode */
                WDSIZE_8    |   /* Transfer Word Size */
                WNR         |   /* DMA Direction */
               nDMAEN       (z);/* DMA Channel Enable */
        w[P5 + lo(DMA7_CONFIG)] = R7;

        IMM32(R7,rx_buf);
        [P5 + lo(DMA7_START_ADDR)] = R7;
    #endif

    call spi_init;


    (R7:6,P5:5) = [SP++];
    unlink;

    rts;
spi_init_rx.end:


/*****************************************************************************
 Function:              spi_init
 Description:           Basic SPI settings, that are not specific to
                        direction and/or master<->slave mode
                        Start SPI
                        Start DMA, if defined
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,R0,P5,ASTAT
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_init:

    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:6,P5:5);
    [--SP] = R0;
    IMM32(P5,SYS_MMR_BASE);


    /* In Master SPI Receive Mode, TIMOD = 10 */
    /* Always enable first DMA, then SPI */
    /* SPI raises immediately an interrupt when enabled */

    /* In Master SPI Transmit Mode, TIMOD = 11 */
    /* SPI starts with the first write access of the DMA */
    /* When SPI is enabled (without enabled DMA), SPI clock is going down (if pulled-up) */


    /******
    * DMA *
    ******/
    #if defined SPI_DMA
        R7 = 0 (z);
        [P5 + lo(DMA7_NEXT_DESC_PTR)] = R7;
        [P5 + lo(DMA7_CURR_DESC_PTR)] = R7;
        [P5 + lo(DMA7_CURR_ADDR)] = R7;
        R7 = PMAP_SPI (z);
        w[P5 + lo(DMA7_PERIPHERAL_MAP)] = R7;
        R7 = DMA7_X_COUNT_VAL (z);
        w[P5 + lo(DMA7_X_COUNT)] = R7;
        R7 = DMA7_Y_COUNT_VAL (z);
        w[P5 + lo(DMA7_Y_COUNT)] = R7;
        R7 = DMA7_X_MODIFY_VAL (z);
        w[P5 + lo(DMA7_X_MODIFY)] = R7;
        R7 = DMA7_Y_MODIFY_VAL (z);
        w[P5 + lo(DMA7_Y_MODIFY)] = R7;

        R7 = w[P5 + lo(DMA7_CONFIG)] (z);

        #if defined SPI_WDSIZE_16
            bitset(R7,bitpos(WDSIZE_16));
        #endif

        w[P5 + lo(DMA7_CONFIG)] = R7;
    #endif


    /******
    * SPI *
    ******/
    #if defined SPI_MASTER_MODE
        #if defined SPI_BAUD_VAL
            R7 = SPI_BAUD_VAL (z);
            /* ensure that SPI baud register value is >= 2 */
            CC = R7 < 2;
            if !CC jump rskip;
            R7 = 2 (z);
            rskip:
            /* set SPI Baud register value */
            w[P5 + lo(SPI_BAUD)] = R7;
        #else
            call spi_calc_baud;
            w[P5 + lo(SPI_BAUD)] = R0;
        #endif

        R7 = w[P5 + lo(SPI_FLG)] (z);
        R6 = 0x1;
        R6 <<= SPI_SLAVE;
        R7 = R7 | R6;
        #if defined SPI_CPHA /* if CPHA = 1 and MSTR = 1, write to SPI_FLG to select slave(s) via FLGx bits */
            R6 <<= 8;
            R6 = ~R6;
            R7 = R6 & R7;
        #endif
        w[P5 + lo(SPI_FLG)] = R7;
    #endif


    R7 = w[P5 + lo(SPI_CTL)] (z);

    #if defined SPI_WDSIZE_16
        bitset(R7,bitpos(SIZE));
    #endif
    #if defined SPI_LSBF
        bitset(R7,bitpos(LSBF));
    #endif
    #if defined SPI_CPHA
        bitset(R7,bitpos(CPHA));
    #endif
    #if defined SPI_CPOL_LOW
        bitset(R7,bitpos(CPOL));
    #endif
    #if defined SPI_MASTER_MODE
        bitset(R7,bitpos(MSTR));
    #endif

    w[P5 + lo(SPI_CTL)] = R7;


    R0 = [SP++];
    (R7:6,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_init.end:


/*****************************************************************************
 Function:              spi_start
 Description:           Basic SPI settings, that are not specific to
                        direction and/or master<->slave mode
                        Start SPI
                        Start DMA, if defined
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7:6,P5
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_start:

    link 0;
    [--SP] = (R7:6,P5:5);
    IMM32(P5,SYS_MMR_BASE);


    /* In Master SPI Receive Mode, TIMOD = 10 */
    /* Always enable first DMA, then SPI */
    /* SPI raises immediately an interrupt when enabled */

    /* In Master SPI Transmit Mode, TIMOD = 11 */
    /* SPI starts with the first write access of the DMA */
    /* When SPI is enabled (without enabled DMA), SPI clock is going down (if pulled-up) */


    /******
    * DMA *
    ******/
    #if defined SPI_DMA
        R7 = w[P5 + lo(DMA7_CONFIG)] (z);
        bitset(R7,bitpos(DMAEN));
        w[P5 + lo(DMA7_CONFIG)] = R7;
        ssync;
    #endif


    /******
    * SPI *
    ******/
    R7 = w[P5 + lo(SPI_CTL)] (z);
    bitset(R7,bitpos(SPE));
    w[P5 + lo(SPI_CTL)] = R7;
    ssync;

    #if defined (SPI_SLAVE_MODE) && defined (PFHW)
        call hostwait_deassert;
    #endif

    #if defined SPI_DMA
        idle; /* wait for SPI DMA interrupt wake-up */
    #endif


    (R7:6,P5:5) = [SP++];
    unlink;

    rts;
spi_start.end:


/****************************************************************************
 Function:              __spi_stop
 Description:           Stop and disable SPI (and SPI DMA)
                        Stop and disable SPI DMA, if defined
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__spi_stop:

    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    #if defined (SPI_SLAVE_MODE) && defined (PFHW)
        call __hostwait_assert;
    #endif

    #if defined SPI_DMA
        R7 = 0 (z);
        w[P5 + lo(DMA7_CONFIG)] = R7; /* Disable DMA */
    #endif


    #if defined SPI_MASTER_MODE
        R7 = 0xFF00 (z);
        w[P5 + lo(SPI_FLG)] = R7; /* reset SPI_FLAG register */
    #endif

    R7 = 0 (z);
    w[P5 + lo(SPI_BAUD)] = R7;

    R7 = CPHA (z);
    w[P5 + lo(SPI_CTL)] = R7;   /* reset SPI_CTL register -> disable SPI */

    R7 =       nSPIF    |   /* SPI Finished (RO) */
                MODF    |   /* Mode Fault Error (W1C) */
                TXE     |   /* Transmission Error (W1C) */
               nTXS     |   /* TDBR Data Buffer Status (RO) */
                RBSY    |   /* Receive Error (W1C) */
               nRXS     |   /* RDBR Data Buffer Status (RO) */
                TXCOL   (z);/* Transmit Collision Error (W1C) */
    w[P5 + lo(SPI_STAT)] = R7; /* W1C: clear status bits */

    R7 = w[P5 + lo(SPI_RDBR)] (z); /* dummy read to clear RXS bit */
    R7 = w[P5 + lo(SPI_SHADOW)] (z); /* dummy read to clear RXS bit */


    (R7:7,P5:5) = [SP++];
    unlink;

    rts;
__spi_stop.end:


#if defined SPI_MASTER_MODE
/****************************************************************************
 Function:              spi_calc_baud
 Description:           generic SPI baud rate calculation function
 Input Parameters:      none
 Return Parameters:     R0
 Registers Used:        R1,R0,ASTAT
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

spi_calc_baud:

    link 0;
    [--SP] = ASTAT;
    [--SP] = R1;

    call __get_sclk_hz;
    IMM32(R1,2*FSPI0CLK_Hz);
    udiv32();

    /* SPI_BAUD <= 0xFFFF */
    spi_calc_baud.check_once:
    IMM32(R1,0xFFFF);
    CC = R1 < R0;
    if !CC jump spi_calc_baud.check_twice;
    R0 = R1;
    jump spi_calc_baud.completed;

    /* SPI_BAUD > 0x1 */
    spi_calc_baud.check_twice:
    R1 = 2 (z);
    CC = R0 < R1;
    if !CC jump spi_calc_baud.completed;
    R0 = R1;

    spi_calc_baud.completed:
    R1 = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
spi_calc_baud.end:
#endif /* SPI_MASTER_MODE */


#if defined PFHW
/*****************************************************************************
 Function:              __hostwait_assert
 Description:           assert hostwait signal, active low signal
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Variables Used: none
 C-Callable :           no
******************************************************************************/

__hostwait_assert:
    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    R7 = PFHW (z);
    w[P5 + lo(PORTFIO_CLEAR)] = R7;
    ssync;

    (R7:7,P5:5) = [SP++];
    unlink;

    rts;
__hostwait_assert.end:


/*****************************************************************************
 Function:              hostwait_deassert
 Description:           de-assert hostwait signal, active low signal
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5
 Global Variables Used: none
 C-Callable :           no
******************************************************************************/

hostwait_deassert:
    link 0;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    R7 = PFHW (z);
    w[P5 + lo(PORTFIO_SET)] = R7;
    ssync;

    (R7:7,P5:5) = [SP++];
    unlink;

    rts;
hostwait_deassert.end:


/*****************************************************************************
 Function:              __hostwait_check
 Description:           wait for HOSTWAIT signal to be deasserted
 Input Parameters:      none
 Return Parameters:     none
 Registers Used:        R7,P5,ASTAT
 Global Variables Used: none
 C-Callable :           no
*****************************************************************************/

__hostwait_check:
    link 0;
    [--SP] = ASTAT;
    [--SP] = (R7:7,P5:5);
    IMM32(P5,SYS_MMR_BASE);

    __hostwait_check.wait:
    R7 = w[P5 + lo(PORTFIO)] (z);
    CC = bittst(R7,bitpos(PFHW));
    if !CC jump __hostwait_check.wait;

    (R7:7,P5:5) = [SP++];
    ASTAT = [SP++];
    unlink;

    rts;
__hostwait_check.END:
#endif /* defined PFHW */


/*****************************************************************************
 EOF
******************************************************************************/
